"""
Commandline interface for making an API request with the SDK.
"""
import argparse
import json
from datetime import datetime
from uuid import UUID
from google.protobuf.json_format import MessageToJson

from cmd import IdentityClient


def main():
    # Create parent parser
    parser = argparse.ArgumentParser(description="Identity client API.")
    parser.add_argument("-l", "--local", action="store_true", help="make the request to localhost")
    subparsers = parser.add_subparsers(dest="command", help="sub-command help")

    # Create child parsers
    introspect_parser = subparsers.add_parser("introspect")
    introspect_parser.add_argument("user_token", help="JWT bearer token")

    verify_parser = subparsers.add_parser("verify")
    verify_parser.add_argument("verification_token", help="Token from email to verify")

    password_change = subparsers.add_parser("change-password")
    password_change.add_argument("user_token", help="JWT bearer token")
    password_change.add_argument("new_password", help="New password for the user in '' (single quotation mark)")

    password_change_of_user = subparsers.add_parser("change-password-of-user")
    password_change_of_user.add_argument("digital_twin_id", help="UUID4 ID of the digital twin for password change")
    password_change_of_user.add_argument("tenant_id", help="UUID4 ID of the tenant")
    password_change_of_user.add_argument("new_password", help="New password for the user in '' (single quotation mark)")

    get_dt = subparsers.add_parser("get-dt")
    get_dt.add_argument("digital_twin_id", help="UUID4 ID of the digital twin for password change")
    get_dt.add_argument("tenant_id", help="UUID4 ID of the tenant")
    get_dt.add_argument("--property_list", nargs="+", help="Array list of the required properties")

    get_dt_by_token = subparsers.add_parser("get-dt-by-token")
    get_dt_by_token.add_argument("user_token", help="JWT bearer token")
    get_dt_by_token.add_argument("--property_list", nargs="+", help="Array list of the required properties")

#     patch_properties = subparsers.add_parser("patch-properties")
#     patch_properties.add_argument("digital_twin_id", help="UUID4 ID of the digital twin for password change")
#     patch_properties.add_argument("tenant_id", help="UUID4 ID of the tenant")
#     patch_properties.add_argument("--add", nargs="+", help="Name and value of the property to add (--add email x@x.x)")
#     patch_properties.add_argument("--add_with_ref", nargs="+", help='''
# Name and value of the property where the value is a reference
#     ''')
#     patch_properties.add_argument("--replace", nargs="+", help="Property ID and new value (--replace 111 a@a.a)")
#     patch_properties.add_argument("--replace_with_ref", nargs="+", help='''
# Property ID and value of the property where the value is a reference
#         ''')
#     patch_properties.add_argument("--remove", nargs="+", help="Remove the properties with the given ID")
#
    patch_properties_by_token = subparsers.add_parser("patch-properties-by-token")
    patch_properties_by_token.add_argument("user_token", help="JWT bearer token")
    patch_properties_by_token.add_argument("--add", nargs="+", help='''
Name and value of the property to add (--add email x@x.x)''')
    patch_properties_by_token.add_argument("--add_by_ref", nargs="+", help='''
Name and value of the property where the value is a reference
        ''')
    patch_properties_by_token.add_argument("--replace", nargs="+", help='''
Property ID and new value (--replace 111 a@a.a)''')
    patch_properties_by_token.add_argument("--replace_by_ref", nargs="+", help='''
Property ID and value of the property where the value is a reference
            ''')
    patch_properties_by_token.add_argument("--remove", nargs="+", help="Remove the properties with the given IDs")

    start_dt_email_verification = subparsers.add_parser("start-dt-email-verification")
    start_dt_email_verification.add_argument("digital_twin", help="UUID4 of the digital twin")
    start_dt_email_verification.add_argument("tenant_id", help="UUID4 of the tenant")
    start_dt_email_verification.add_argument("email", help="email address to validate")

    args = parser.parse_args()

    local = args.local
    client = IdentityClient(local)

    command = args.command

    if command == "introspect":
        user_token = args.user_token
        token_info = client.introspect_token(user_token)
        if token_info is not None:
            print_token_info(token_info)

    elif command == "verify":
        verification_token = args.verification_token
        digital_twin_info = client.verify_digital_twin_email(verification_token)
        if digital_twin_info is not None:
            print_verify_info(digital_twin_info)

    elif command == "change-password":
        user_token = args.user_token
        new_password = args.new_password
        password_change = client.change_password(user_token, new_password)
        if password_change is not None:
            print(password_change)

    elif command == "change-password-of-user":
        digital_twin_id = args.digital_twin_id
        tenant_id = args.tenant_id
        new_password = args.new_password
        password_change = client.change_password_of_user(digital_twin_id, tenant_id, new_password)
        if password_change is not None:
            print(password_change)

    elif command == "get-dt":
        digital_twin_id = args.digital_twin_id
        tenant_id = args.tenant_id
        property_list = args.property_list
        dt = client.get_digital_twin(digital_twin_id, tenant_id, property_list)
        if dt is not None:
            print(dt)

    elif command == "get-dt-by-token":
        user_token = args.user_token
        property_list = args.property_list
        dt = client.get_digital_twin_by_token(user_token, property_list)
        if dt is not None:
            print(dt)

    # elif command == "patch-properties":
    #     digital_twin_id = args.digital_twin_id
    #     tenant_id = args.tenant_id
    #     all_args = {
    #         "add": [],
    #         "add_with_ref": [],
    #         "replace": [],
    #         "replace_with_ref": [],
    #         "remove": []
    #     }

        # props = args.add
        # add_args_to_dict(all_args, "add", props)
        # props = args.add_by_ref
        # add_args_to_dict(all_args, "add_with_ref", props)
        # props = args.replace
        # add_args_to_dict(all_args, "replace", props)
        # props = args.replace_by_ref
        # add_args_to_dict(all_args, "replace_with_ref", props)
        # props = args.remove
        # add_args_to_dict(all_args, "remove", props)
        # properties = client.patch_properties(digital_twin_id, tenant_id, all_args)
        # if properties is not None:
        #     print(properties)

    elif command == "patch-properties-by-token":
        user_token = args.user_token
        all_args = {
            "add": [],
            "add_by_ref": [],
            "replace": [],
            "replace_by_ref": [],
            "remove": []
        }

        props = args.add
        add_args_to_dict(all_args, "add", props)
        props = args.add_by_ref
        add_args_to_dict(all_args, "add_by_ref", props)
        props = args.replace
        add_args_to_dict(all_args, "replace", props)
        props = args.replace_by_ref
        add_args_to_dict(all_args, "replace_by_ref", props)
        props = args.remove
        add_args_to_dict(all_args, "remove", props)
        properties = client.patch_properties_by_token(user_token, all_args)
        if properties is not None:
            print(properties)

    elif command == "start-dt-email-verification":
        digital_twin_id = args.digital_twin
        tenant_id = args.tenant_id
        email = args.email
        resp = client.start_digital_twin_email_verification(digital_twin_id, tenant_id, email)
        if resp is not None:
            print(resp)


def print_verify_info(digital_twin_info):
    print("Digital twin info")
    print("=================")
    print("Tenant: " + str(UUID(bytes=digital_twin_info.digital_twin.tenant_id)))
    print("Digital twin: " + str(UUID(bytes=digital_twin_info.digital_twin.id)))


def print_token_info(token_info):
    print("Token info")
    print("==========")
    print("Tenant: " + str(UUID(bytes=token_info.tenant_id)))
    print("Customer: " + str(UUID(bytes=token_info.customer_id)))
    print("App space: " + str(UUID(bytes=token_info.app_space_id)))
    print("Application: " + str(UUID(bytes=token_info.application_id)))
    print("Subject: " + str(UUID(bytes=token_info.subject_id)))
    print("Expire time: " + str(datetime.fromtimestamp(token_info.expire_time.seconds)))


def print_response(resp):
    js = MessageToJson(resp)
    js_dict = json.loads(js)
    # print("----: ", end="")
    # print(type(js_dict["digitalTwin"]["properties"]), end=" | ")
    # print(type(js_dict["digitalTwin"]["digitalTwin"]["id"]))
    # print(list(js_dict))
    prettify(js_dict)
    print(js_dict)
    # outer_keys = list(js_dict)
    # for i in outer_keys:
    #     if isinstance(i, type(dict)):
    #         prettify(i)


def dump_object(obj):
    for descriptor in obj.DESCRIPTOR.fields:
        value = getattr(obj, descriptor.name)
        # print(descriptor.name)
        if descriptor.label == descriptor.LABEL_REPEATED:
            map(dump_object, value)
        if descriptor.type == descriptor.TYPE_MESSAGE:
            dump_object(value)
        # elif descriptor.type == descriptor.TYPE_ENUM:
        #     enum_name = descriptor.enum_type.values[value].name
        #     print("%s: %s" % (descriptor.full_name, enum_name))
        elif "id" in descriptor.name:
            setattr(obj, descriptor.name, byte_to_uuid(value))
            # print("%s: %s" % (descriptor.full_name, byte_to_uuid(value)))

def prettify(js):
    # print(type(js))
    for k, v in js.items():
        if isinstance(v, type(dict())):
            # print(k)
            prettify(v)
        elif isinstance(v, type(list())):
            # print("\t" + k, end=".")
            for val in v:
                if isinstance(val, type(str())):
                    js[k] = byte_to_uuid(v)
                    pass
                elif isinstance(val, type(list())):
                    pass
                else:
                    prettify(val)
        else:
            if isinstance(v, str):
                js[k] = byte_to_uuid(v)
            # else:
            #     x = v
            # print("\t" + str(k) + ': ' + str(x))


def byte_to_uuid(value):
    try:
        # x = value.decode("hex")
        uid = UUID(bytes=value)
    except ValueError:
        return value
    return uid


def add_args_to_dict(all_args, action, values):
    if action == "add" and values is not None:
        for v in values:
            all_args["add"].append(v)
    elif action == "add_by_ref" and values is not None:
        for v in values:
            all_args["add_by_ref"].append(v)
    elif action == "replace" and values is not None:
        for v in values:
            all_args["replace"].append(v)
    elif action == "replace_by_ref" and values is not None:
        for v in values:
            all_args["replace_by_ref"].append(v)
    elif action == "remove" and values is not None:
        for v in values:
            try:
                identity = int(v)
            except TypeError:
                print("The given value: " + v + " is not an ID")
            all_args["remove"].append(identity)

    return all_args


if __name__ == '__main__':
    main()
